﻿Imports System.Windows.Forms


Class WaveFilePlayer

    Private Const mVolumeScaling As Single = 1000.0F

    ' ----------------------------------------------------- LOAD
    Private mLoadSamplesCount As Int32
    Private mLoadWaveTable() As Single

    ' ----------------------------------------------------- OPERATION
    Private mSamplesCount As Int32
    Private mWaveTable() As Single

    Private mWaveFilePosition As Int32
    Private mPositionNormalized As Double

    Private mFreq As Single = 0
    Private mFreqOld As Single
    Private mSamplesPerSec As Int32 = 44100

    Private mPan As Single
    Private mVolumeUP As Single
    Private mVolumeDW As Single
    Private mVolume_New As Int32
    Private mVolume_Actual As Int32

    Private mFadeIn As Boolean = False
    Private mFadeOut As Boolean = False
    Private mReachedTheEnd As Boolean = False

    ' -----------------------------------------------------
    Private mDesiredPos As Single
    Private mMargin As Single
    Private mResponseSpeed As Single
    Private mMaxSpeedUP As Single
    Private mMaxSpeedDW As Single

    ' ----------------------------------------------------- buffer
    Private mBufferSize As Int32
    Private mBufferCount As Int32
    Private mPhase As Int64 = 0
    Private mPlayBuffer() As Single
    Private mBufferFillTimeMicrosec As Int32

    Private mTriggerMode As Boolean = False
    Private mHiQuality As Boolean = True

    Private mFollow As Boolean = True
    Private mLooping As Boolean = False

    Private mOutDevice As Int32 = -1
    Private mStereo As Boolean = False
    Private mNumChannels As Int32 = 1

    Private mFileName As String = ""

    ' ----------------------------------------------------- player
    Private mPlayer As AudioOut = New AudioOut

    ' ----------------------------------------------------- effects
    Friend Effects As Effects = New Effects


    ' ------------------------------------------------------------------------- WAVE FILE
    Private Sub LoadWaveFile()

        Dim fs As WaveStream = Nothing

        Try

            Dim MediaPath As String = Application.StartupPath & "\Media\"

            fs = New WaveStream(MediaPath & mFileName)

            Dim byteArray(CInt(fs.Length) - 1) As Byte

            fs.Read(byteArray, 0, CInt(fs.Length))

            If fs.Format.wBitsPerSample <> 16 Then
                MsgBox("The file """ & mFileName & """ is not 16 bit per sample." & vbCrLf & vbCrLf & _
                        "Please use only the format: Wave - PCM - Signed - 16 bit - Stereo or Mono - at 44100 samples per second." & vbCrLf, _
                        MsgBoxStyle.Exclamation)
                End
            End If

            mSamplesPerSec = fs.Format.nSamplesPerSec

            Dim bh1 As Byte
            Dim bl1 As Byte
            Dim bh2 As Byte
            Dim bl2 As Byte
            Dim v1 As Single
            Dim v2 As Single

            If mNumChannels = 1 Then
                If fs.Format.nChannels = 1 Then
                    ' ----------------------------------------------------------- mono to mono
                    mLoadSamplesCount = CInt(fs.Length \ 2)
                    ReDim mLoadWaveTable(mLoadSamplesCount - 1)
                    For i As Int32 = 0 To mLoadSamplesCount - 1
                        bl1 = byteArray(i * 2)
                        bh1 = byteArray(i * 2 + 1)
                        If bh1 > 127 Then
                            v1 = (CInt(bh1 - 256) * 256 + bl1) / 32768.0F
                        Else
                            v1 = (CInt(bh1) * 256 + bl1) / 32768.0F
                        End If
                        mLoadWaveTable(i) = v1
                    Next
                Else
                    ' ----------------------------------------------------------- stereo to mono
                    mLoadSamplesCount = CInt(fs.Length \ 4)
                    ReDim mLoadWaveTable(mLoadSamplesCount * 2 - 1)
                    For i As Int32 = 0 To mLoadSamplesCount - 1
                        bl1 = byteArray(i * 4)
                        bh1 = byteArray(i * 4 + 1)
                        bl2 = byteArray(i * 4 + 2)
                        bh2 = byteArray(i * 4 + 3)
                        If bh1 > 127 Then
                            v1 = (CInt(bh1 - 256) * 256 + bl1) / 32768.0F
                        Else
                            v1 = (CInt(bh1) * 256 + bl1) / 32768.0F
                        End If
                        If bh2 > 127 Then
                            v2 = (CInt(bh2 - 256) * 256 + bl2) / 32768.0F
                        Else
                            v2 = (CInt(bh2) * 256 + bl2) / 32768.0F
                        End If
                        mLoadWaveTable(i) = (v1 + v2) / 2
                    Next
                End If
            Else
                If fs.Format.nChannels = 1 Then
                    ' ----------------------------------------------------------- mono to stereo
                    mLoadSamplesCount = CInt(fs.Length \ 2)
                    ReDim mLoadWaveTable(mLoadSamplesCount * 2 - 1)
                    For i As Int32 = 0 To mLoadSamplesCount - 1
                        bl1 = byteArray(i * 2)
                        bh1 = byteArray(i * 2 + 1)
                        If bh1 > 127 Then
                            v1 = (CInt(bh1 - 256) * 256 + bl1) / 32768.0F
                        Else
                            v1 = (CInt(bh1) * 256 + bl1) / 32768.0F
                        End If
                        mLoadWaveTable(i * 2) = v1
                        mLoadWaveTable(i * 2 + 1) = v1
                    Next
                Else
                    ' ----------------------------------------------------------- stereo to stereo
                    mLoadSamplesCount = CInt(fs.Length \ 4)
                    ReDim mLoadWaveTable(mLoadSamplesCount * 2 - 1)
                    For i As Int32 = 0 To mLoadSamplesCount - 1
                        bl1 = byteArray(i * 4)
                        bh1 = byteArray(i * 4 + 1)
                        bl2 = byteArray(i * 4 + 2)
                        bh2 = byteArray(i * 4 + 3)
                        If bh1 > 127 Then
                            v1 = (CInt(bh1 - 256) * 256 + bl1) / 32768.0F
                        Else
                            v1 = (CInt(bh1) * 256 + bl1) / 32768.0F
                        End If
                        If bh2 > 127 Then
                            v2 = (CInt(bh2 - 256) * 256 + bl2) / 32768.0F
                        Else
                            v2 = (CInt(bh2) * 256 + bl2) / 32768.0F
                        End If
                        mLoadWaveTable(i * 2) = v1
                        mLoadWaveTable(i * 2 + 1) = v2
                    Next
                End If
            End If

        Catch
            ReDim mLoadWaveTable(-1)
            MsgBox("Can not load the file """ & mFileName & """" & vbCrLf, MsgBoxStyle.Exclamation)
        Finally
            If fs IsNot Nothing Then fs.Dispose()
        End Try

    End Sub

    'Friend Sub LoadWaveFile_WithThread()
    '    ' --------------------------------------------------------- asynchronously start the Thread 
    '    Dim objThread As Threading.Thread
    '    objThread = New Threading.Thread(AddressOf LoadWaveFile)
    '    ' --------------------------------------------------------- make the thread as background thread.
    '    objThread.IsBackground = True
    '    ' --------------------------------------------------------- set the Priority of the thread.
    '    objThread.Priority = Threading.ThreadPriority.BelowNormal
    '    ' --------------------------------------------------------- start the thread.
    '    objThread.Start()
    'End Sub



    ' ------------------------------------------------------------------------------------------------
    '  FillBuffer - Int64 VERSION - Sample smoothing at low speed - Looping or Limited
    ' ------------------------------------------------------------------------------------------------
    '   9uS to fill 1024 samples
    '   max audio len = 16 hours ( 44100 Hz )
    ' ------------------------------------------------------------------------------------------------
    Private Const ShiftBits As Byte = 24 ' ( shiftbits max = 30 )
    ' ------------------------------------------------------------------------------------------------
    Private Sub FillBuffer(ByRef data() As Int16)

        ' -------------------------------------------------------------------------- SyncLock allows to change mWaveTable
        SyncLock Me
            Try
                Dim t As PrecisionTimer = New PrecisionTimer

                ' ---------------------------------------------------------------------- get size (16 bit samples count)
                Dim size As Int32 = data.Length

                ' ---------------------------------------------------------------------- redim buffer if needed
                If mPlayBuffer Is Nothing OrElse mPlayBuffer.Length <> size Then
                    ReDim mPlayBuffer(size - 1)
                End If

                ' ---------------------------------------------------------------------- prepare freq & phase values
                Dim nSamples As Int32 = size \ mNumChannels
                Dim pmax As Int64 = CLng(mSamplesCount - mNumChannels) << ShiftBits
                Dim f As Int64 = CLng(mFreqOld * (1 << ShiftBits))
                Dim df As Int64 = CLng((mFreq - mFreqOld) * (1 << ShiftBits) / nSamples)
                mFreqOld = mFreq

                ' ---------------------------------------------------------------------- prepare Volume values
                Dim DeltaVol As Int32 = 0
                If mFadeOut Then
                    mVolume_New = 0
                    DeltaVol = -1
                Else
                    If f > 0 Then
                        mVolume_New = CInt(mVolumeUP * mVolumeScaling)
                    Else
                        mVolume_New = CInt(mVolumeDW * mVolumeScaling)
                    End If

                    ' -----------------------------------------------------------------------
                    ' TODO - test if the uS time is not increased with zero volume
                    ' -----------------------------------------------------------------------
                    ' Volume = 0 generates minimum representations of IEEE floating-points, 
                    ' results are likely "denormal" that are sometimes processed more slowly. 
                    ' -----------------------------------------------------------------------
                    ' This correction was removed the 2015 03 07 
                    ' If mVolume_New < 1 Then mVolume_New = 1
                    ' -----------------------------------------------------------------------

                    ' -----------------------------------------------------------------------
                    If mVolume_New > mVolume_Actual Then
                        DeltaVol = 1
                    Else
                        DeltaVol = -1
                    End If
                End If


                ' ---------------------------------------------------------------------- select QUALITY and STEREO
                Dim type As Int32 = 0
                If mHiQuality Or mFadeIn Or mFadeOut Then type += 1
                If mNumChannels = 2 Then type += 4

                ' ---------------------------------------------------------------------- prepare for smoothing


                Dim valueNew As Single
                Static valueOld As Single = 0

                Dim valueNewLeft As Single
                Dim valueNewRight As Single
                Static valueOldLeft As Single = 0
                Static valueOldRight As Single = 0


                ' ----------------------------------------------------- SMOOTING
                '   freq    freq        smoothing(not sq)  smoothing
                '   1       44.1KHz     ---                ---
                '   0.4     17.6KHz     1                  1 
                '   0.2      8.8KHz     2                  4  
                '   0.1      4.4KHz     4                  16  
                '   0.01     441 Hz     40                 1600 
                '   0.004    176 Hz     100                10000 
                ' ---------------------------------------------------------------
                Dim smoothing As Single
                Dim absFreq As Single = Math.Abs(mFreq)
                If mHiQuality Or mFadeIn Or mFadeOut Then
                    If absFreq <= 0.01F Then
                        ' -------------------------------------------- very low speed 
                        type += 2
                    Else
                        If absFreq < 0.4F Then
                            ' ---------------------------------------- low speed 
                            type += 1
                            smoothing = CSng(0.4F / absFreq)
                            If smoothing < 1 Then smoothing = 1
                            smoothing = smoothing * smoothing
                        End If
                    End If
                End If

                ' -------------------------------------------------------------------------- FILL BUFFER
                Select Case type

                    Case 0 ' MONO - LowQuality
                        For i As Int32 = 0 To nSamples - 1
                            ' --------------------------------------------------
                            mPlayBuffer(i) = mWaveTable(CInt(mPhase >> ShiftBits)) * mVolume_New / mVolumeScaling
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 1
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase < 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next
                        valueOld = mPlayBuffer(nSamples - 1)

                    Case 1 ' MONO - HiQuality - HiSpeed
                        For i As Int32 = 0 To nSamples - 1
                            ' --------------------------------------------------
                            mPlayBuffer(i) = mWaveTable(CInt(mPhase >> ShiftBits)) * mVolume_Actual / mVolumeScaling
                            ' -------------------------------------------------- apply volume changes
                            If mVolume_New <> mVolume_Actual Then mVolume_Actual += DeltaVol
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 1
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase <= 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next
                        valueOld = mPlayBuffer(nSamples - 1)

                    Case 2 ' MONO - HiQuality - LowSpeed
                        For i As Int32 = 0 To nSamples - 1
                            ' --------------------------------------------------
                            valueNew = mWaveTable(CInt(mPhase >> ShiftBits)) * mVolume_Actual / mVolumeScaling
                            ' -------------------------------------------------- apply volume changes
                            If mVolume_New <> mVolume_Actual Then mVolume_Actual += DeltaVol
                            ' -------------------------------------------------- apply the low-speed smoothing
                            valueOld += (valueNew - valueOld) / smoothing

                            mPlayBuffer(i) = valueOld
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 1
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase <= 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next

                    Case 3 ' MONO - HiQuality - VeryLowSpeed
                        For i As Int32 = 0 To nSamples - 1
                            mPlayBuffer(i) = valueOld
                        Next

                    Case 4 ' STEREO - LowQuality
                        Dim ix As Int32
                        Dim VolumeLeft As Single
                        Dim VolumeRight As Single
                        If mPan > 0 Then
                            VolumeLeft = mVolume_New * (1 - mPan)
                            VolumeRight = mVolume_New
                        Else
                            VolumeLeft = mVolume_New
                            VolumeRight = mVolume_New * (1 + mPan)
                        End If
                        For i As Int32 = 0 To nSamples - 1
                            ' --------------------------------------------------
                            ix = 2 * CInt(mPhase >> ShiftBits)
                            mPlayBuffer(i * 2) = mWaveTable(ix) * VolumeLeft / mVolumeScaling
                            mPlayBuffer(i * 2 + 1) = mWaveTable(ix + 1) * VolumeRight / mVolumeScaling
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 2
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase < 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next
                        valueOldLeft = mPlayBuffer(2 * (nSamples - 1))
                        valueOldRight = mPlayBuffer(2 * (nSamples - 1) + 1)

                    Case 5 ' STEREO - HiQuality - HiSpeed
                        Dim ix As Int32
                        For i As Int32 = 0 To nSamples - 1
                            ' -------------------------------------------------- 
                            Dim VolumeLeft As Single = mVolume_Actual
                            Dim VolumeRight As Single = mVolume_Actual
                            If mPan > 0 Then
                                VolumeLeft = VolumeLeft * (1 - mPan)
                            Else
                                VolumeRight = VolumeRight * (1 + mPan)
                            End If
                            ' --------------------------------------------------
                            ix = 2 * CInt(mPhase >> ShiftBits)
                            mPlayBuffer(i * 2) = mWaveTable(ix) * VolumeLeft / mVolumeScaling
                            mPlayBuffer(i * 2 + 1) = mWaveTable(ix + 1) * VolumeRight / mVolumeScaling
                            ' -------------------------------------------------- apply volume changes
                            If mVolume_New <> mVolume_Actual Then mVolume_Actual += DeltaVol
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 2
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase < 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next
                        valueOldLeft = mPlayBuffer(2 * (nSamples - 1))
                        valueOldRight = mPlayBuffer(2 * (nSamples - 1) + 1)

                    Case 6 ' STEREO - HiQuality - LowSpeed
                        Dim ix As Int32
                        For i As Int32 = 0 To nSamples - 1
                            ' --------------------------------------------------
                            Dim VolumeLeft As Single = mVolume_Actual
                            Dim VolumeRight As Single = mVolume_Actual
                            If mPan > 0 Then
                                VolumeLeft = VolumeLeft * (1 - mPan)
                            Else
                                VolumeRight = VolumeRight * (1 + mPan)
                            End If
                            ' --------------------------------------------------
                            ix = 2 * CInt(mPhase >> ShiftBits)
                            valueNewLeft = mWaveTable(ix) * VolumeLeft / mVolumeScaling
                            valueNewRight = mWaveTable(ix + 1) * VolumeRight / mVolumeScaling
                            ' -------------------------------------------------- apply volume changes
                            If mVolume_New <> mVolume_Actual Then mVolume_Actual += DeltaVol
                            ' -------------------------------------------------- apply the low-speed smoothing
                            valueOldLeft += (valueNewLeft - valueOldLeft) / smoothing
                            valueOldRight += (valueNewRight - valueOldRight) / smoothing

                            mPlayBuffer(i * 2) = valueOldLeft
                            mPlayBuffer(i * 2 + 1) = valueOldRight
                            ' -------------------------------------------------- increment the phase
                            mPhase += f
                            f += df
                            ' -------------------------------------------------- test limits
                            If mPhase >= pmax Then
                                If mLooping Then
                                    mPhase -= pmax
                                Else
                                    mPhase = pmax - 2
                                End If
                                mReachedTheEnd = True
                            End If
                            If mPhase < 0 Then
                                If mLooping Then
                                    mPhase += pmax
                                Else
                                    mPhase = 0
                                End If
                                mReachedTheEnd = True
                            End If
                        Next

                    Case 7 ' STEREO - HiQuality - VeryLowSpeed
                        For i As Int32 = 0 To nSamples - 1
                            mPlayBuffer(i * 2) = valueOldLeft
                            mPlayBuffer(i * 2 + 1) = valueOldRight
                        Next

                End Select

                mWaveFilePosition = CInt(mPhase >> ShiftBits)
                mPositionNormalized = mWaveFilePosition / mSamplesCount

                'Debug.Print(CInt(mphase >> ShiftBits).ToString)

                ' ---------------------------------------------------------------------- effects
                Effects.Synthesize(mPlayBuffer, mStereo)

                ' -----------------------------------------------------------  Todo - Remove - Nan Correction
                'If Single.IsNaN(mPlayBuffer(0)) Then Return
                ' -----------------------------------------------------------

                ' ---------------------------------------------------------------------- limit and convert
                Dim v As Single
                For i As Int32 = 0 To nSamples * mNumChannels - 1
                    v = mPlayBuffer(i)
                    ' ------------------------------------------------------------------ hard limit. +1/-1
                    'v = 0.5F * (Math.Abs(v + 1.0F) - Math.Abs(v - 1.0F))
                    If v > 1 Then v = 1
                    If v < -1 Then v = -1
                    ' ------------------------------------------------------------------ soft limit
                    'v = 1.5F * v - 0.5F * v * v * v
                    ' ------------------------------------------------------------------ convert to SHORT in the output array
                    data(i) = CShort(v * 32767.0F)
                Next

                mBufferFillTimeMicrosec = CInt(t.GetTimeMicrosec)

            Catch
            End Try
        End SyncLock
    End Sub




    Friend Sub TimedUpdate()

        If Not mTriggerMode Then
            Dim CurrentPos As Double = PositionNormalized()

            ' ----------------------------------------------------------------------- calculate desired frequency
            Dim NewSpeed As Single
            Static OldSpeed As Single

            If mFollow Then
                Dim desired As Double = mDesiredPos
                If mMargin > 0 Then
                    desired += -mMargin + desired * mMargin * 2.0
                End If
                Dim delta As Double = desired - CurrentPos
                NewSpeed = CSng(mResponseSpeed * mSamplesCount * 0.0005 * delta)
                OldSpeed += (NewSpeed - OldSpeed) * mResponseSpeed
            Else
                If mDesiredPos > 0.5 Then
                    NewSpeed = mMaxSpeedUP * 2.0F * (mDesiredPos - 0.5F)
                Else
                    NewSpeed = -mMaxSpeedDW * 2.0F * (0.5F - mDesiredPos)
                End If
                ' -------------------------------------------------------------------
                OldSpeed += (NewSpeed - OldSpeed) * mResponseSpeed
            End If

            ' ----------------------------------------------------------------------- limit desired frequency
            If OldSpeed > mMaxSpeedUP Then OldSpeed = mMaxSpeedUP
            If OldSpeed < -mMaxSpeedDW Then OldSpeed = -mMaxSpeedDW

            mFreq = OldSpeed
        End If


        ' --------------------------------------------------------------------------- FadeIn and FadeOut
        Static DelayCounter As Int32 = 0
        If mFadeIn Then
            DelayCounter += 1
            If DelayCounter > 8 Then
                DelayCounter = 0
                mFadeIn = False
            End If
        End If
        If mFadeOut Then
            DelayCounter += 1
            If DelayCounter > 12 Then
                DelayCounter = 0
                mFadeOut = False
            End If
        End If

    End Sub



    ' =================================================================================================
    '  PUBLIC PROPERTIES
    ' =================================================================================================

    Friend Property ReachedTheEnd() As Boolean
        Get
            Return mReachedTheEnd
        End Get
        Set(ByVal value As Boolean)
            mReachedTheEnd = value
        End Set
    End Property

    Friend ReadOnly Property BufferSamplesCount() As Int32
        Get
            Return mBufferSize
        End Get
    End Property

    Friend ReadOnly Property BufferFillTimeMicrosec() As Int32
        Get
            If mPlayer Is Nothing Then Return 0
            Return mBufferFillTimeMicrosec
        End Get
    End Property

    Friend ReadOnly Property TriggerMode() As Boolean
        Get
            Return mTriggerMode
        End Get
    End Property

    Friend ReadOnly Property HiQuality() As Boolean
        Get
            Return mHiQuality
        End Get
    End Property

    Friend Property Follow() As Boolean
        Get
            Return mFollow
        End Get
        Set(ByVal value As Boolean)
            mFollow = value
        End Set
    End Property

    Friend Property Looping() As Boolean
        Get
            Return mLooping
        End Get
        Set(ByVal value As Boolean)
            mLooping = value
        End Set
    End Property

    Friend Property OutDevice() As Int32
        Get
            Return mOutDevice
        End Get
        Set(ByVal value As Int32)
            mOutDevice = value
        End Set
    End Property

    Friend ReadOnly Property Stereo() As Boolean
        Get
            Return mStereo
        End Get
    End Property

    Friend Property ResponseSpeed() As Int32
        Get
            Return CInt(mResponseSpeed * 100.0F)
        End Get
        Set(ByVal value As Int32)
            mResponseSpeed = value / 100.0F
        End Set
    End Property

    Friend Property Margin() As Int32
        Get
            Return CInt(mMargin * 100.0F)
        End Get
        Set(ByVal value As Int32)
            mMargin = value / 100.0F
        End Set
    End Property

    Friend Property MaxSpeedUP() As Double
        Get
            Return mMaxSpeedUP
        End Get
        Set(ByVal value As Double)
            mMaxSpeedUP = CSng(value)
        End Set
    End Property

    Friend Property MaxSpeedDW() As Double
        Get
            Return mMaxSpeedDW
        End Get
        Set(ByVal value As Double)
            mMaxSpeedDW = CSng(value)
        End Set
    End Property

    Friend ReadOnly Property PositionNormalized() As Double
        Get
            Return mPositionNormalized
        End Get
    End Property

    Friend Property Pan() As Single
        Get
            Return mPan
        End Get
        Set(ByVal value As Single)
            If value > 1 Then value = 1
            If value < -1 Then value = -1
            mPan = value
        End Set
    End Property

    Friend Property VolumeUP() As Single
        Get
            Return CSng(Math.Sqrt(mVolumeUP))
        End Get
        Set(ByVal value As Single)
            If value > 1 Then value = 1
            If value < 0 Then value = 0
            mVolumeUP = value * value
        End Set
    End Property

    Friend Property VolumeDW() As Single
        Get
            Return CSng(Math.Sqrt(mVolumeDW))
        End Get
        Set(ByVal value As Single)
            If value > 1 Then value = 1
            If value < 0 Then value = 0
            mVolumeDW = value * value
        End Set
    End Property

    Friend Property Frequency() As Single
        Get
            Return mFreq
        End Get
        Set(ByVal value As Single)
            mFreq = value
        End Set
    End Property

    Friend WriteOnly Property DesiredPos() As Single
        Set(ByVal value As Single)
            mDesiredPos = value
        End Set
    End Property

    Friend ReadOnly Property FileName() As String
        Get
            Return mFileName
        End Get
    End Property

    Friend Sub PlayPause()
        mFreq = 0
        mFreqOld = 0
    End Sub

    Friend Sub PlayStop()
        Form1.Timer1.Stop()
        ' ------------------------------------ fade out
        mFadeOut = True
        mFreq = 0
        mFreqOld = 0
        ' ---------------------------------------- 
        For i As Int32 = 1 To 2
            TimedUpdate()
            Form1.UpdateWavePositionBar()
            System.Threading.Thread.Sleep(10)
        Next
        ' ---------------------------------------- close player
        If mPlayer IsNot Nothing Then
            mPlayer.OutputOff()
        End If
    End Sub


    Friend Sub PlayStart(ByVal FileName As String, _
                         ByVal Stereo As Boolean, _
                         ByVal TriggerMode As Boolean, _
                         ByVal HiQuality As Boolean, _
                         ByVal Looping As Boolean)

        If FileName = "" Then Return
        If mOutDevice >= WaveNative.OutputDevicesCount Then Return

        ' ---------------------------------------- SyncLock allows to change mWaveTable
        SyncLock Me

            Form1.Timer1.Stop()

            If FileName <> mFileName Or Stereo <> mStereo Then
                mFileName = FileName
                If Stereo Then
                    mNumChannels = 2
                    mStereo = True
                Else
                    mNumChannels = 1
                    mStereo = False
                End If
                LoadWaveFile()
                'LoadWaveFile_WithThread()
            End If

            ' -------------------------------------------------------------- STOP
            PlayPause()

            mTriggerMode = TriggerMode
            mHiQuality = HiQuality
            mLooping = Looping

            If Not TriggerMode Then
                mFreq = 0
            End If

            ' -------------------------------------------------------------- set corresponding position for the new file
            If mSamplesCount > 0 Then
                Try
                    mPhase = (mPhase \ mSamplesCount) * mLoadSamplesCount
                Catch
                End Try
            End If

            ' -------------------------------------------------------------- copy the array
            mSamplesCount = mLoadSamplesCount
            ReDim mWaveTable(mLoadWaveTable.Length - 1)
            Array.Copy(mLoadWaveTable, mWaveTable, mSamplesCount * mNumChannels)

        End SyncLock

        ' -------------------------------------------------------------- start new sound
        mFadeIn = True
        mBufferCount = 8
        mBufferSize = 1024 * mNumChannels
        Dim sps As Int32 = 44100 ' max audiocard rate normally = 192000
        sps = mSamplesPerSec
        mPlayer.OutputOn(mOutDevice, sps, mNumChannels, mBufferSize, mBufferCount, AddressOf FillBuffer)

        Form1.Timer1.Start()

    End Sub

End Class
